import re
import sys
import gmpy2
import base64
import hashlib

from functools import reduce

from OpenSSL import crypto

from pyasn1.type.univ import Sequence, SequenceOf, Integer
from pyasn1.type.namedtype import NamedType, NamedTypes, OptionalNamedType
from pyasn1.codec.der.encoder import encode as der_encode

def product(items):
    return reduce(lambda x, y: x * y, items)

class RSAError(ValueError):
    pass

# https://tools.ietf.org/html/rfc8017#appendix-A.1.2
class OtherPrimeInfo(Sequence):
    componentType = NamedTypes(
        NamedType('prime', Integer()),
        NamedType('exponent', Integer()),
        NamedType('coefficient', Integer())
    )

class OtherPrimeInfos(SequenceOf):
    componentType = OtherPrimeInfo()

class RSAPrivateKey(Sequence):
    componentType = NamedTypes(
        # Version ::= INTEGER { two-prime(0), multi(1) }
        NamedType('version', Integer()),
        NamedType('modulus', Integer()),
        NamedType('publicExponent', Integer()),
        NamedType('privateExponent', Integer()),
        NamedType('prime1', Integer()),
        NamedType('prime2', Integer()),
        NamedType('exponent1', Integer()),
        NamedType('exponent2', Integer()),
        NamedType('coefficient', Integer()),
        OptionalNamedType('otherPrimeInfos', OtherPrimeInfos())
    )

    # instantiate private key given public exponent and list of primes
    def __init__(self, e, primes):
        super().__init__()

        if len(set(primes)) != len(primes):
            raise RSAError('Cannot compute CRT coefficents with repeat primes')

        primes = [2] + primes

        # compute modulus
        n = product(primes[1:])

        primes_minus_one = [x - 1 for x in primes]
        phi = product(primes_minus_one[1:])

        if gmpy2.gcd(e, phi) != 1:
            raise RSAError('e and phi(n) are not coprime')

        # compute private exponent
        d = int(gmpy2.invert(e, phi))

        exponents = [d % x for x in primes_minus_one]

        coefficients = [None] * len(primes)
        # inv q mod p
        coefficients[2] = int(gmpy2.invert(primes[2], primes[1]))
        for i in range(3, len(primes)):
            # inv product(r[1] ... r[i-1]) mod r[i]
            coefficients[i] = int(gmpy2.invert(product(primes[1:i]), primes[i]))

        # base set of rsa private key values
        self['modulus'] = n
        self['publicExponent'] = e
        self['privateExponent'] = d
        self['prime1'] = primes[1]
        self['prime2'] = primes[2]
        self['exponent1'] = exponents[1]
        self['exponent2'] = exponents[2]
        self['coefficient'] = coefficients[2]

        # add values for additional primes if needed
        if (len(primes) - 1) > 2:
            self['version'] = 1
            otherPrimeInfos = OtherPrimeInfos()
            for i in range(3, len(primes)):
                info = OtherPrimeInfo()
                info['prime'] = primes[i]
                info['exponent'] = exponents[i]
                info['coefficient'] = coefficients[i]
                otherPrimeInfos.setComponentByPosition(i - 3, info)

            self['otherPrimeInfos'] = otherPrimeInfos
        else:
            self['version'] = 0

    @property
    def der(self):
        return der_encode(self)

    @property
    def pem(self):
        b64 = base64.b64encode(self.der).decode()
        b64 = '\n'.join(b64[i:i+64] for i in range(0, len(b64), 64))
        return '-----BEGIN {text}-----\n{b64}\n-----END {text}-----\n'.format(
            b64=b64,
            text='RSA PRIVATE KEY'
        )

if __name__ == '__main__':
    e = int(sys.argv[1])
    primes = list(map(int, sys.argv[2:]))
    sys.stdout.write(RSAPrivateKey(e, primes).pem)
